#!/bin/bash

set -e

INSTPREFIX="$1"

NODE=node

if [ -x /usr/bin/node ]; then
    NODE=/usr/bin/node
elif [ -x /usr/bin/nodejs ]; then
    NODE=/usr/bin/nodejs
else
    echo "WARNING: node not found"
fi

echo $NODE
$NODE --version

VER=8.2.0.1
WORKDIR=${PWD}
TOPDIR0=/opt/ghcjs/8.2.0.1/lib/ghcjs

BINDIR=${INSTPREFIX}/opt/ghcjs/8.2.0.1/bin
TOPDIR=${INSTPREFIX}${TOPDIR0}

if [ ! -f "${PWD}/cabal.rc" ]; then
    echo "cabal.rc not found in pwd ($PWD)"
    exit 1
fi

mkdir -p /tmp/cabal

rm -rf "$INSTPREFIX"

PATH=/opt/ghc/8.2.2/bin:/opt/happy/1.19.5/bin:/opt/alex/3.1.7/bin:$PATH HOME=/tmp/home CABAL_CONFIG="${PWD}/cabal.rc" \
    /opt/cabal/head/bin/cabal --project-file=./cabal.ghcjs.project new-build :pkg:ghcjs

rm -fv .ghc.environment.x86_64-linux-8.2.2

mkdir -p $TOPDIR/bin
for N in ghcjs ghcjs-pkg haddock-ghcjs hsc2hs-ghcjs; do
    cp -v $PWD/dist-newstyle/build/x86_64-linux/ghc-8.2.2/ghcjs-8.2.0.1/build/$N/$N $TOPDIR/bin/$N
done

GHCJSBOOT=$PWD/dist-newstyle/build/x86_64-linux/ghc-8.2.2/ghcjs-8.2.0.1/build/ghcjs-boot/ghcjs-boot
GHCJSRUN=$PWD/dist-newstyle/build/x86_64-linux/ghc-8.2.2/ghcjs-8.2.0.1/build/ghcjs-run/ghcjs-run
BOOTAR=$PWD/ghcjs-8.2.0.1/data/boot.tar


############################################################################
## building core binaries phase done; here starts boot library bootstrap phase

mkdir -p "${BINDIR}"

# temporary inplace wrappers
for N in ghcjs ghcjs-pkg haddock-ghcjs hsc2hs-ghcjs; do
    sed "s|@TOPDIR@|$TOPDIR|g" "./wrappers/$N.in" > "$BINDIR/$N"
    chmod +x "$BINDIR/$N"
done


HOME=/tmp/cabal CABAL_CONFIG="${PWD}/cabal.rc" PATH="${BINDIR}:/opt/ghc/8.2.2/bin:$PATH" \
    $GHCJSBOOT -s $BOOTAR \
    --no-prof --no-haddock \
    --with-node $NODE \
    --with-ghcjs-run $GHCJSRUN \
    --with-cabal /opt/cabal/head/bin/cabal

# fixup installation                                                                                                                                                                                                                                           

## fixup database                                                                                                                                                                                                                                              
sed -i "s|$TOPDIR|\${pkgroot}|g" "${TOPDIR}/package.conf.d"/*.conf
"$BINDIR/ghcjs-pkg" recache -v

## remove redundant stuff                                                                                                                                                                                                                                      
rm -rf "$TOPDIR/boot"

## create proper wrappers
for N in ghcjs ghcjs-pkg haddock-ghcjs hsc2hs-ghcjs; do
    rm -fv "$BINDIR/$N"
    sed "s|@TOPDIR@|$TOPDIR0|g" "./wrappers/$N.in" > "$BINDIR/$N-$VER"
    chmod +x "$BINDIR/$N-$VER"
done

exit 0
